import pytest

from mhctyper.hla_allele import (
    HLAllele,
    HLAllelePattern,
    decompose,
    reduce_resolution,
)


@pytest.fixture(autouse=True, scope="module")
def hla_allele():
    return HLAllele


@pytest.fixture(autouse=True, scope="module")
def hla_allele_default_pattern():
    return HLAllelePattern()


def test_hlallele_default_pattern():
    ap = HLAllelePattern()
    assert ap.prefix == "(?P<prefix>(:?HLA|hla)[-_])"
    assert ap.locus == "(?P<locus>[a-zA-Z]+[0-9]?)"
    assert (
        ap.digit_fields
        == "(?P<digit_fields>(?:[0-9]+[:_]){0,3}[0-9NLSCAQnlscaq]+)"
    )


@pytest.mark.parametrize("resolution, expect", [(-1, 1), (5, 4)])
def test_hlaallele_resolution_value(resolution, expect):
    """
    Test if HLAllelePattern.resolution is set correctly,
    when its value is either less than 1 or greater than 4
    """
    ap = HLAllelePattern(resolution=resolution)
    assert ap.resolution == expect


def test_hlaallele_digit_field_with_custom_resolution():
    ap = HLAllelePattern(resolution=2)
    assert (
        ap.digit_fields
        == "(?P<digit_fields>(?:[0-9]+[:_]){0,1}[0-9NLSCAQnlscaq]+)"
    )
    ap = HLAllelePattern(resolution=0)
    assert (
        ap.digit_fields
        == "(?P<digit_fields>(?:[0-9]+[:_]){0,0}[0-9NLSCAQnlscaq]+)"
    )
    ap = HLAllelePattern(resolution=7)
    assert (
        ap.digit_fields
        == "(?P<digit_fields>(?:[0-9]+[:_]){0,3}[0-9NLSCAQnlscaq]+)"
    )


def test_empty_pattern():
    with pytest.raises(ValueError):
        HLAllelePattern(locus="", prefix="   ")


def test_pattern_with_empty_whitespaces():
    """
    Test if whitespaces are trimmed off from ends when there are
    whitespaces in the given pattern
    """
    ap = HLAllelePattern(locus="    ([A-Z]+[0-9]) ")
    assert ap.locus == "(?P<locus>([A-Z]+[0-9]))"
    ap = HLAllelePattern(locus="\t([A-Z]+[0-9])\n")
    assert ap.locus == "(?P<locus>([A-Z]+[0-9]))"


@pytest.mark.parametrize(
    "allele, nth_field, expect",
    [
        ("A*02:06", 1, "A*02"),
        ("hla_a_01_01_01", 2, "hla_a_01_01"),
        ("hla_drb1*11_01_01:347N", 3, "hla_drb1*11_01_01"),
        ("HLA-dqb1*06:02", 1, "HLA-dqb1*06"),
        ("hla_drb1*11_01_01:347N", -1, "hla_drb1*11"),
    ],
)
def test_reduce_resolution(allele, nth_field, expect, recwarn):
    ap = HLAllelePattern(resolution=nth_field)
    reduced_allele = reduce_resolution(allele, ap)
    assert len(recwarn) == 0
    assert reduced_allele == expect


@pytest.mark.parametrize(
    "allele, nth_field",
    [
        ("A*02", 1),
        ("hla_a_01_01_01", 3),
        ("hla_drb1*11_01_01:347N", 4),
        ("HLA-dqb1*06:02", 3),
    ],
)
def test_reduce_resolution_warning(allele, nth_field):
    with pytest.warns(
        RuntimeWarning, match="No resolution reduced for input allele="
    ):
        ap = HLAllelePattern(resolution=nth_field)
        reduce_resolution(allele, ap)


@pytest.mark.parametrize(
    "allele, nth_field, expect",
    [
        ("hla_a_01-01-01-01", 2, "hla_a_01-01"),
        ("DQB1*11-01-01-01", 2, "DQB1*11-01"),
        ("HLA-B*11:01-01_01", 3, "HLA-B*11:01-01"),
    ],
)
def test_reduce_resolution_with_custom_pattern(allele, nth_field, expect):
    """
    Test reduce resolution with custom pattern
    """
    # Allow digit field using either -, :, _, or a mixture of 3 as separator
    ap = HLAllelePattern(digit_field_sep=["-", ":", "_"], resolution=nth_field)
    assert reduce_resolution(allele, ap) == expect


def test_decompose_empty_allele(hla_allele_default_pattern):
    with pytest.raises(ValueError):
        decompose("", hla_allele_default_pattern)


@pytest.mark.parametrize(
    "allele",
    [
        "jsldkfjlasdjwe",
        "A*",
        "HLA",
        "HLA-C*",
        "hla_drb1_",
        "hla_dqb1_02_01_01_",
        "hla_c*01-01-01",
        "hla_c-01-01-01",
        "_01_01_01",
        "*11:01:01",
        "01-01-01",
        "11:01:01",
        "11_01_01",
    ],
)
def test_decompose_fail_to_find_given_pattern(
    allele, hla_allele_default_pattern
):
    with pytest.raises(ValueError):
        reduce_resolution(allele, hla_allele_default_pattern)
        decompose(allele, hla_allele_default_pattern)


@pytest.mark.parametrize(
    "allele, expect",
    [
        ("A*01:01:01", "A"),
        ("HLA-c*01:03", "c"),
        ("HLA-B*04:16N", "B"),
        ("HLA-A*02:376N", "A"),
        ("hla_dqb1_01_01_01", "dqb1"),
        ("hla_drb1_11_01_01_01", "drb1"),
    ],
)
def test_decompose_locus(allele, expect, hla_allele_default_pattern):
    assert decompose(allele, hla_allele_default_pattern).locus == expect


@pytest.mark.parametrize(
    "allele",
    [
        ("QQQ*01:01:01"),
        ("HLA-ivy*01:03"),
        ("hla_drq1_01_01_01"),
    ],
)
def test_decompose_bad_locus(allele, hla_allele_default_pattern):
    with pytest.raises(ValueError):
        decompose(allele, hla_allele_default_pattern)


@pytest.mark.parametrize(
    "allele, expect",
    [
        ("A*01:01:01", "01:01:01"),
        ("HLA-B*04:16N", "04:16N"),
        ("hla-c*02:376n", "02:376n"),
        ("hla_c_07", "07"),
    ],
)
def test_decompose_digit_field(allele, expect, hla_allele_default_pattern):
    assert decompose(allele, hla_allele_default_pattern).digit_field == expect


def test_hlallele_str():
    """Test str(HLAllele())"""
    assert (
        str(
            HLAllele(prefix="", locus="C", digit_field="01:01:01:02n", sep="*")
        )
        == "C*01:01:01:02n"
    )
    assert (
        str(
            HLAllele(
                prefix="hla_",
                locus="DRB1",
                digit_field="01_01_01:02n",
                sep="_",
            )
        )
        == "hla_DRB1_01_01_01:02n"
    )



from __future__ import annotations

import sys
import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING

import polars as pl
from tinyscibio import BAMetadata

from .hla_allele import HLAllelePattern, reduce_resolution
from .logger import logger

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Optional

    from tinyscibio import _PathLike


def load_rg_sm_from_bam(bam_metadata: BAMetadata) -> str:
    logger.info("Load SM value from read group in the given BAM.")
    try:
        rg = bam_metadata.read_groups
        logger.debug(f"bam_metadata.read_groups: {rg=}")
        if len(rg) > 1:
            raise ValueError(f"Found more than 1 read groups: {rg}")
        rg_sm = next(iter(rg)).get("SM", "")
        logger.debug(f"{rg_sm=}")
        if not rg_sm:
            raise ValueError(
                "Failed to get SM from read group. "
                f"Please check parsed read group: {rg}"
            )
        return rg_sm
    except ValueError as e:
        logger.error(e)
        sys.exit(1)


def load_allele_pop_freq(freq_fspath: _PathLike) -> pl.DataFrame:
    """Load allele population frequency data"""
    logger.info("Load HLA alleles from population frequency file.")
    freq_df = pl.read_csv(freq_fspath, separator="\t")
    # remove supertype has all zero pop frequency
    return freq_df.filter(
        pl.fold(0, lambda acc, s: acc + s, pl.all().exclude(pl.String)) > 0.0
    )


def collect_alleles_to_type(
    bam_metadata: BAMetadata,
    kept: Optional[Sequence[str]] = None,
) -> list[str]:
    """Collect HLA alleles to type"""
    logger.info("Load HLA alleles to type.")
    try:
        ap = HLAllelePattern(resolution=2)
        alleles_df = pl.DataFrame({"Allele": bam_metadata.seqnames()})
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if kept is not None:
                alleles_df = alleles_df.with_columns(
                    pl.col("Allele")
                    .map_elements(
                        lambda c: reduce_resolution(c, ap),
                        return_dtype=pl.String,
                    )
                    .alias("D4")
                ).filter(pl.col("D4").is_in(kept))
        alleles = alleles_df["Allele"].to_list()
        if not alleles:
            raise ValueError("Failed to collect any alleles for typing")
        logger.info(f"Loaded {len(alleles)} alleles to type.")
        return alleles
    except ValueError as e:
        logger.error(e)
        sys.exit(1)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Optional

    from tinyscibio import _PathLike


class Logger(logging.Logger):
    def __init__(self, name: str):
        super().__init__(name)
        self._configured = False

    def initialize(
        self, debug: bool = False, f: Optional[_PathLike] = None
    ) -> None:
        if self._configured:
            return

        self.setLevel(logging.DEBUG if debug else logging.INFO)
        if self.hasHandlers():
            self.handlers.clear()

        console_handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "[%(module)s.py] - [%(funcName)s:%(lineno)d] - [%(levelname)s]: %(message)s"
        )
        console_handler.setFormatter(formatter)
        console_handler.setLevel(logging.INFO)
        self.addHandler(console_handler)

        if debug:
            if f is None:
                raise ValueError(
                    f"Parameter {f=} is required to add a FileHandler to "
                    "logger when debug is true."
                )
            file_handler = logging.FileHandler(f, mode="a")
            file_handler.setFormatter(formatter)
            file_handler.setLevel(logging.DEBUG)
            self.addHandler(file_handler)

        self._configured = True


logging.setLoggerClass(Logger)
# logging.getLogger returns Logger type
# Although setLoggerClass(Logger) sets the logger object as Logger class,
# the getLogger method, according to typeshed, still returns Logger type
# Thats why LSP (e.g. pyright) complains about initialize is not an attribute
# https://github.com/microsoft/pyright/discussions/5418
logger = cast(Logger, logging.getLogger(__name__))



import logging
import math
import sys
from functools import partial
from multiprocessing import get_context
from typing import Optional, cast

import numpy as np
import polars as pl
from tinyscibio import BAMetadata, walk_bam
from tqdm import tqdm

from .hla_allele import HLAllelePattern, decompose
from .logger import logger


def score_log_liklihood(
    row: dict[str, list[int] | list[str]], scale: float = math.exp(23)
) -> float:
    score: float = 0.0
    start, end = 0, 0
    base_qs = cast(list[int], row["bqs"])
    md = cast(list[str], row["mds"])
    for i in range(len(md)):
        if md[i].startswith("^"):
            continue
        block = np.array([])
        if md[i].isalpha():
            # mismatches
            end = start + len(md[i])
            block = np.array([10 ** (-k / 10) / 3 for k in base_qs[start:end]])
        else:
            # matches
            end = start + int(md[i])
            block = np.array([1 - 10 ** (-k / 10) for k in base_qs[start:end]])
        score += np.sum(np.log(np.multiply(block, scale)))
        start = end
    return score


def score_per_allele(
    allele: str,
    bam_fspath: str,
    min_ecnt: int,
    log_fspath: Optional[str] = None,
) -> pl.DataFrame | None:
    debug = True if log_fspath is not None else False
    logger.initialize(debug, log_fspath)

    ap = HLAllelePattern()
    hla_allele = decompose(allele, ap)
    logger.debug(f"{hla_allele=}")
    hla_gene = f"{hla_allele.prefix}{hla_allele.locus}"
    logger.debug(f"{hla_gene=}")

    bametadata = BAMetadata(bam_fspath)
    df = walk_bam(
        bam_fspath,
        allele,
        exclude=3584,
        return_ecnt=True,
        return_bq=True,
        return_md=True,
        return_qname=True,
    )
    # walk_bam returns rname as id
    # convert it back to rname
    df = df.with_columns(
        pl.col("rnames").replace_strict(bametadata.idx2seqname()),
    )
    logger.debug(f"walk_bam returns {df.shape[0]} alignments.")
    logger.debug(
        f"walk_bam return {df.filter(~ pl.col('propers')).shape[0]} "
        "non-proper alignments."
    )
    logger.debug(
        f"walk_bam return {df.filter(pl.col('indel_ecnt') > 0).shape[0]} "
        "alignments with indels."
    )
    logger.debug(
        f"walk_bam return {df.filter(pl.col('mm_ecnt') > min_ecnt).shape[0]} "
        "alignments with mms more than allowed min_ecnt."
    )
    df = df.filter(
        (pl.col("indel_ecnt") == 0)
        & (pl.col("mm_ecnt") <= min_ecnt)  # good aln
        & (pl.col("propers"))
    )
    logger.debug(
        f"{df.shape[0]} alignments left after filtering "
        "for proper, indel_ecnt, and mm_enct."
    )
    # we only keep read in pairs after applying above filters
    # do not do the following before the above
    df = (
        df.with_columns(
            pl.col("qnames").count().over("qnames").alias("n"),
        )
        .filter(pl.col("n") == 2)
        .drop("n")
    )
    logger.debug(f"{df.shape[0]} alignments left after filtering for paired.")
    if df.shape[0] == 0:
        logger.debug("no alignments left for scoring after filtering. Return")
        return None
    # we convert dtype of bqs and mds column to get ready for score likelihood
    df = df.with_columns(
        pl.col("bqs").map_elements(
            lambda x: x.tolist(),
            return_dtype=pl.List(pl.UInt8),
        ),
        pl.col("mds").map_elements(
            lambda x: x, return_dtype=pl.List(pl.String)
        ),
    )
    # score the likelihood given bqs and mds
    df = df.with_columns(
        pl.struct(["bqs", "mds"])
        .map_elements(score_log_liklihood, return_dtype=pl.Float64)
        .alias("scores"),
    )
    # Sum up the score per aligned pair
    df = df.group_by("qnames").agg(pl.col("scores").sum())
    logger.debug(f"After score and sum per pair: {df}")
    df = df.with_columns(allele=pl.lit(allele), gene=pl.lit(hla_gene))
    return df


def score_a_one(
    alleles_to_score: list[str],
    bam: str,
    min_ecnt: int,
    nproc: int = 8,
    debug: bool = False,
) -> pl.DataFrame:
    # gets the file handler
    log_fspath = None
    if debug:
        log_fh = next(
            iter(
                [
                    h
                    for h in logger.handlers
                    if isinstance(h, logging.FileHandler)
                ]
            )
        )
        log_fspath = log_fh.baseFilename
        # set nproc to 1 when in debug mode
        logger.debug("Debug mode: setting nproc to 1.")
        nproc = 1
    try:
        logger.info("Score first allele.")
        score_tables: list[pl.DataFrame] = []
        logger.debug(f"# alleles to score: {len(alleles_to_score)}.")
        with get_context("spawn").Pool(processes=nproc) as pool:
            task_iterator = tqdm(
                pool.imap_unordered(
                    partial(
                        score_per_allele,
                        bam_fspath=bam,
                        min_ecnt=min_ecnt,
                        log_fspath=log_fspath,  # pass to child proc
                    ),
                    alleles_to_score,
                ),
                total=len(alleles_to_score),
                desc="Score first allele: ",
                ncols=100,  # define width
            )
            for res in task_iterator:
                if res is None:
                    continue
                # display allele being processed
                task_iterator.set_postfix(
                    {"allele": f"{res['allele'].unique().item()}"}
                )
                score_tables.append(res)
        if not score_tables:
            raise ValueError("Failed to score for any first alleles.")
        scores = pl.concat([s for s in score_tables])
        logger.info(f"Alleles scored: {len(score_tables)}.")
    except ValueError as e:
        logger.error(e)
        sys.exit(1)
    return scores


def score_second_by_gene(
    gene: str, a1_scores: pl.DataFrame, a1_winners: pl.DataFrame
) -> pl.DataFrame:
    a1_winners = a1_winners.filter(pl.col("gene") == gene)
    a1_scores = a1_scores.filter(pl.col("gene") == gene)
    score_table = a1_scores.join(a1_winners, on=["qnames", "gene"], how="left")
    score_table = score_table.with_columns(
        pl.col("scores_right").fill_null(0.0)
    )
    score_table = score_table.with_columns(
        factor=pl.col("scores") / (pl.col("scores") + pl.col("scores_right"))
    )
    score_table = score_table.with_columns(
        scores=pl.col("scores") * pl.col("factor")
    )
    return score_table


def score_a_two(
    a1_scores: pl.DataFrame,
    a1_winners: pl.DataFrame,
    nproc: int = 8,
) -> pl.DataFrame:
    logger.info("Score second allele.")
    try:
        score_tables: list[pl.DataFrame] = []
        genes = a1_winners["gene"].unique().to_list()
        nproc = min(nproc, len(genes))
        with get_context("spawn").Pool(processes=nproc) as pool:
            gene_iterator = tqdm(
                pool.imap_unordered(
                    partial(
                        score_second_by_gene,
                        a1_scores=a1_scores,
                        a1_winners=a1_winners,
                    ),
                    genes,
                ),
                total=len(genes),
                desc="Score second allele: ",
                ncols=100,
            )
            for res in gene_iterator:
                if res is None:
                    continue
                score_tables.append(res)
        if not score_tables:
            raise ValueError("Failed to score for any second alleles.")
        a2_scores = pl.concat(score_tables)
        return a2_scores
    except ValueError as e:
        logger.error(e)
        sys.exit(1)


def get_winners(allele_scores: pl.DataFrame) -> pl.DataFrame:
    # round scores to 4 decimal places to avoid precision
    # problem when getting alleles whose scores equal to max scores
    tot_scores = allele_scores.group_by(["allele", "gene"]).agg(
        pl.col("scores").sum().round(4)
    )
    winners = tot_scores.filter(
        pl.col("scores") == pl.col("scores").max().over("gene")
    )
    logger.debug("winner alleles who have the max score per locus.")
    if winners.shape[0] > 3:
        logger.debug("there are ties in scores for certain gene groups.")
    logger.debug(winners)
    # when there is a tie in scores for each gene group,
    # select allele with least string value lexicographically
    # e.g. hla_a_26_01_24 and hla_a_26_01_01 (latter selected)
    winners = winners.sort(by=["allele"]).unique(subset="gene", keep="first")
    winners = winners.rename({"scores": "tot_scores"})
    return winners



import argparse

from tinyscibio import parse_path


def parse_cmd() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--bam",
        metavar="FILE",
        type=parse_path,
        required=True,
        help="specify path to BAM file",
    )
    parser.add_argument(
        "--freq",
        metavar="FILE",
        type=parse_path,
        required=True,
        help="specify path to HLA frequency file.",
    )
    parser.add_argument(
        "--outdir",
        metavar="DIR",
        type=parse_path,
        required=True,
        help="specify path to output folder.",
    )
    parser.add_argument(
        "--min_ecnt",
        metavar="INT",
        type=int,
        default=999,
        help="specify minimum # of mm events (999).",
    )
    parser.add_argument(
        "--nproc",
        metavar="INT",
        type=int,
        default=8,
        help="specify # processes to use (8).",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="specify to overwrite scores."
    )
    parser.add_argument(
        "--debug", action="store_true", help="specify to enter debug mode."
    )
    return parser



#!/usr/bin/env python

from __future__ import annotations

import polars as pl
from tinyscibio import BAMetadata, get_parent_dir, make_dir

from .cli import parse_cmd
from .logger import logger
from .score_alleles import get_winners, score_a_one, score_a_two
from .utils import (
    collect_alleles_to_type,
    load_allele_pop_freq,
    load_rg_sm_from_bam,
)


def run_mhctyper() -> int:
    parser = parse_cmd()
    args = parser.parse_args()

    make_dir(args.outdir, exist_ok=True, parents=True)

    # set up logger accordingly
    if args.debug:
        debug_log_fspath = args.outdir / f"{__name__}.debug.log"
        if debug_log_fspath.exists():
            debug_log_fspath.unlink()
        logger.initialize(args.debug, debug_log_fspath)
    else:
        logger.initialize(args.debug)

    logger.info(f"Start HLA typing from given BAM file: {args.bam}")

    allele_pop_freq = load_allele_pop_freq(freq_fspath=args.freq)

    bam_metadata = BAMetadata(args.bam)
    # collect all alleles to type from BAM header
    alleles_to_type = collect_alleles_to_type(
        bam_metadata, kept=allele_pop_freq["Allele"].to_list()
    )

    rg_sm = load_rg_sm_from_bam(bam_metadata)

    out_a1 = args.outdir / f"{rg_sm}.a1.tsv"
    out_a2 = args.outdir / f"{rg_sm}.a2.tsv"
    hla_res = args.outdir / f"{rg_sm}.hlatyping.res.tsv"
    if args.overwrite:
        logger.info("Overwrite specified. Delete results previously computed.")
        out_a1.unlink(missing_ok=True)
        out_a2.unlink(missing_ok=True)
        hla_res.unlink(missing_ok=True)

    a1_scores = pl.DataFrame()
    if not out_a1.exists():
        a1_scores = score_a_one(
            alleles_to_score=alleles_to_type,
            bam=args.bam,
            min_ecnt=args.min_ecnt,
            nproc=args.nproc,
            debug=args.debug,
        )
        a1_scores.write_csv(out_a1, separator="\t")
    else:
        logger.info("Found scores of first alleles previously computed.")
        a1_scores = pl.read_csv(out_a1, separator="\t")

    logger.info("Get winner for the first typed allele.")
    a1_winners = get_winners(allele_scores=a1_scores)
    winner_scores = a1_scores.join(
        a1_winners, on=["gene", "allele"], how="inner"
    )

    a2_scores = score_a_two(
        a1_scores=a1_scores,
        a1_winners=winner_scores,
        nproc=args.nproc,
    )
    a2_scores.write_csv(out_a2, separator="\t")
    logger.info("Get winner for the second typed allele.")
    a2_winners = get_winners(allele_scores=a2_scores)

    logger.info("Combine winnes for both first and second alleles.")
    hla_res_df = pl.concat([a1_winners, a2_winners])
    hla_res_df = hla_res_df.with_columns(sample=pl.lit(rg_sm)).sort(
        by="allele"
    )
    logger.info(f"Final HLA typing result: {hla_res_df}")
    hla_res_df.write_csv(hla_res, separator="\t")
    return 0



import re
import warnings
from dataclasses import dataclass, field, fields

__VALID_LOCI: list[str] = [
    "A",
    "B",
    "C",
    "E",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    "N",
    "P",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "DRA",
    "DRB1",
    "DRB2",
    "DRB3",
    "DRB4",
    "DRB5",
    "DRB6",
    "DRB7",
    "DRB8",
    "DRB9",
    "DQA1",
    "DQB1",
    "DQA2",
    "DQB2",
    "DQB3",
    "DOA",
    "DOB",
    "DMA",
    "DMB",
    "DPA1",
    "DPB1",
    "DPA2",
    "DPB2",
    "DPA3",
]


@dataclass
class HLAllelePattern:
    prefix: str = field(default="(:?HLA|hla)[-_]")
    locus: str = field(default="[a-zA-Z]+[0-9]?")
    digit_fields: str = field(init=False)
    digit_field_sep: list[str] = field(default_factory=lambda: [":", "_"])
    expr_suffix: list[str] = field(
        default_factory=lambda: ["N", "L", "S", "C", "A", "Q"]
    )
    resolution: int = field(default=4)
    sep: str = field(default="[\\*_]")

    def __post_init__(self) -> None:
        self._validate_resolution()
        # the first line gives "(?:[0-9]+[:_])"
        # the second line gives {0, resolution-1}
        # the last 2 lines give [0-9NLSCAQnlscaq]+
        self.digit_fields = (
            f"(?:[0-9]+[{''.join(self.digit_field_sep)}])"
            f"{{0,{self.resolution-1}}}"
            f"[0-9{''.join(self.expr_suffix)}"
            f"{''.join([c.lower() for c in self.expr_suffix])}]+"
        )
        self._validate_fields()
        self._add_name_to_group_match()

    def _validate_resolution(self) -> None:
        """Make sure resolution is set between 1 and 4"""
        if self.resolution < 1:
            self.resolution = 1
        if self.resolution > 4:
            self.resolution = 4

    def _validate_fields(self) -> None:
        for f in fields(self):  # f is field object of this class
            v = getattr(self, f.name)
            if isinstance(v, str):
                if not v or not v.strip():
                    raise ValueError(
                        f"Property {f.name} cannot be empty or "
                        "just whitespaces. Whitespaces at beginning and "
                        "end of the pattern string are stripped"
                    )

    def _add_name_to_group_match(self) -> None:
        for f in fields(self):
            v = getattr(self, f.name)
            if isinstance(v, str):
                setattr(self, f.name, f"(?P<{f.name}>{v.strip()})")


@dataclass
class HLAllele:
    """
    HLAllele object to represent a HLA allele

    A HLA allele = "{prefix}{locus}{sep}{fields}"
    """

    prefix: str
    locus: str
    digit_field: str
    sep: str

    def __str__(self) -> str:
        return f"{self.prefix}{self.locus}{self.sep}{self.digit_field}"


def _check_if_allele_empty(allele: str) -> None:
    """Check if given allele string is empty"""
    if not allele or not allele.strip():
        raise ValueError(
            "allele cannot be empty or just whitespaces. "
            "Whitespaces at beginning and end of the pattern "
            "string are stripped"
        )


def reduce_resolution(allele: str, ap: HLAllelePattern) -> str:
    """Reduce resolution of input HLA allele"""
    pattern = f"{ap.prefix}?{ap.locus}{ap.sep}{ap.digit_fields}"

    pattern_c = re.compile(pattern)
    m = pattern_c.search(allele)
    if m is None:
        raise ValueError("Failed to decompose the given allele")

    if m.group() == allele:
        warnings.warn(
            f"No resolution reduced for input {allele=}", RuntimeWarning
        )
    return m.group()


def decompose(allele: str, ap: HLAllelePattern) -> HLAllele:
    """Decompose given HLA allele string"""
    _check_if_allele_empty(allele)

    pattern = f"^{ap.prefix}?{ap.locus}{ap.sep}{ap.digit_fields}$"

    pattern_c = re.compile(pattern)
    m = pattern_c.search(allele)
    # quit when no match found
    if m is None:
        raise ValueError("Failed to decompose the given allele")

    prefix = m.group("prefix") if m.group("prefix") is not None else ""

    locus = m.group("locus")
    # make sure identified locus is valid
    if locus is not None and locus.upper() not in __VALID_LOCI:
        raise ValueError

    return HLAllele(
        prefix=prefix,
        locus=m.group("locus"),
        digit_field=m.group("digit_fields"),
        sep=m.group("sep"),
    )
